import 'package:collection/collection.dart';
import 'package:oobium_datastore_gen/src/class.dart';

class ClassField {

  final String name;
  final String type;
  final bool _nullable;
  final String? initializer;
  final metadata = <String>[];

  ClassField({
    required this.name,
    required this.type,
    bool nullable = false,
    String? initializer,
    Iterable<String>? metadata
  }) :
    _nullable = nullable,
    initializer = initializer {
    metadata?.forEach((m) => this.metadata.add(m));
  }
  ClassField.paramOf(ClassField f) :
        parent = f.parent, type = f.parameter, name ='', _nullable = false, initializer = null;

  late final Class parent;

  String get nullable => isNullable ? '?' : '';

  bool get isDateTime => type == 'DateTime';
  bool get isNotDateTime => !isDateTime;
  bool get isNullable => _nullable;
  bool get isNotNullable => !isNullable;
  bool get isString => type == 'String';
  bool get isNotString => !isString;
  bool get isIterable => type == 'Iterable' || type.startsWith('Iterable<');
  bool get isNotIterable => !isIterable;
  bool get isCollection => isList || isIterable || isSet || isMap;
  bool get isNotCollection => !isCollection;
  bool get isList => type == 'List' || type.startsWith('List<');
  bool get isNotList => !isList;
  bool get isSet => type == 'Set' || type.startsWith('Set<');
  bool get isNotSet => !isSet;
  bool get isMap => type == 'Map' || type.startsWith('Map<');
  bool get isNotMap => !isMap;
  bool get isObjectId => type == 'ObjectId';
  bool get isNotObjectId => !isObjectId;
  bool get isPrimitive => isBool || isInt || isNum || isDouble || isString;
  bool get isNotPrimitive => !isPrimitive;
  bool get isBool => type == 'bool';
  bool get isNotBool => !isBool;
  bool get isInt => type == 'int';
  bool get isNotInt => !isInt;
  bool get isNum => type == 'num';
  bool get isNotNum => !isNum;
  bool get isDouble => type == 'double';
  bool get isNotDouble => !isDouble;
  bool get isImported => parent.schema.imports.any((m) => m.names.any((n) => n == type));
  bool get isNotImported => !isImported;
  bool get isInPart => isNotDateTime && isNotCollection && isNotPrimitive && parent.schema.isInPart(type);
  bool get isNotInPart => !isInPart;
  bool get isEnum => isNotDateTime && isNotCollection && isNotPrimitive && parent.schema.isEnum(type);
  bool get isNotEnum => !isEnum;

  bool get isInitialized => initializer != null;
  bool get isNotInitialized => !isInitialized;

  bool get isParameterized => type.contains('<');
  bool get isNotParameterized => !isParameterized;
  String get parameter => type.substring(type.indexOf('<') + 1, type.length-1);

  bool get isRequired => isNotNullable && isNotInitialized;
  bool get isNotRequired => !isRequired;

  String? meta(String key) => metadata.firstWhereOrNull((m) => m.startsWith('$key:'))?.substring('$key:'.length).trim();

  String? get jsonDecode => meta('jsonDecode');
  String? get jsonEncode => meta('jsonEncode');

  @override
  String toString() => '$name $type$nullable';
}
