import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

String genPersist(Schema schema) => PersistFeatureGenerator(schema).compile();

class PersistFeatureGenerator {

  final Schema schema;
  PersistFeatureGenerator(this.schema);

  PersistFeature get feature => schema.features.whereType<PersistFeature>().first;
  Iterable<Model> get models => schema.models;

  String compileCreator() =>
'''
  _PersistFeature(
    path: path,
    version: ${feature.version},
    compactionStrategy: const DefaultCompactionStrategy(),
    converter: const _Converter(),
    onUpgrade: onUpgrade ?? dropOnUpgrade
  )
''';

  String compile() =>
'''
${schema.partOfDeclaration}

class _PersistFeature extends PersistFeature {
  _PersistFeature({
    required String path,
    required int version,
    required CompactionStrategy compactionStrategy,
    required Converter converter,
    required OnUpgrade onUpgrade
  }) : super(path, version, compactionStrategy, converter, onUpgrade);

  @override
  void putRecord(DataRecord record) {
    final v = converter.decodePut(record);${models.map((m) =>'''
    if(v is ${m.name}) put(v);''').join(' else ')}
    else throw Exception('unknown type \$v during PersistFeature.putRecord');
  }

  @override
  void removeRecord(DataRecord record) {
    final k = converter.decodeRemove(record);${models.map((m) =>'''
    if(k is ${m.keyName}) remove(k);''').join(' else ')}
    else throw Exception('unknown type \$k during PersistFeature.removeRecord');
  }
}
''';
}
