import 'package:oobium_datastore_gen/src/generators/rest_feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class RestAdapterGenerator {

  final Model m;
  Schema get schema => m.schema;
  RestAdapterGenerator(this.m);

  String compile() =>
m.isSingleton ?
'''
class ${m.adapterName} extends _RestAdapter {

  FutureOr<Map<String, dynamic>?> Function($onRequestParams)? _onRequest;
  FutureOr<void> Function($onResultsParams)? _onResults;
  FutureOr<void> Function($onChangeParams)? _onCreated;
  FutureOr<void> Function($onUpdateParams)? _onUpdated;

  ${m.adapterName}({
    FutureOr<Map<String, dynamic>?> Function($onRequestParams)? onRequest,
    FutureOr<void> Function($onResultsParams)? onResults,
    FutureOr<void> Function($onChangeParams)? onCreated,
    FutureOr<void> Function($onUpdateParams)? onUpdated
  }) :
    _onRequest = onRequest,
    _onResults = onResults,
    _onCreated = onCreated,
    _onUpdated = onUpdated;
  
  FutureOr<Map<String, dynamic>?> onRequest($onRequestParams) => _onRequest?.call($onRequestArgs);
  FutureOr<void> onResults($onResultsParams) => _onResults?.call($onResultsArgs);
  FutureOr<void> onCreated($onChangeParams) => _onCreated?.call($onChangeArgs);
  FutureOr<void> onUpdated($onUpdateParams) => _onUpdated?.call($onUpdateArgs);

  FutureOr<void> _created(Executor ex, $_onChangeParams) => onCreated($onChangeArgs);
  FutureOr<void> _updated(Executor ex, $_onUpdateParams) => onUpdated($onUpdateArgs);
  
  void expire() => _fetched = null;
  
  Future<void>? _fetched;
  Future<void> fetch($fetchParams) {
    if (force) {
      final current = _fetched;
      if(current != null) {
        _fetched = null;
        return current.then((_) => fetch($fetchArgs));
      }
    }
    return (_fetched ??= _fetch($_fetchArgs));
  }
  Future<void> _fetch($_fetchParams) async {
    _log.debug(() => 'fetching ${m.name}');
    final data = await onRequest($onRequestArgs);
    if(data == null) {
      _log.debug(() => 'fetched ${m.name}: not found');
    } else {
      await onResults($onResultsArgs);
      final ${m.varName} = ${m.name}.fromJson($jsonArgs);
      _feature.put(${m.varName});
      _log.debug(() => 'fetched \$${m.varName}');
    }
  }
}
''' :
'''
class ${m.adapterName} extends _RestAdapter {
 
  FutureOr<dynamic> Function($onRequestParams)? _onRequest;
  FutureOr<void> Function($onResultsParams)? _onResults;
  FutureOr<void> Function($onChangeParams)? _onCreated;
  FutureOr<void> Function($onUpdateParams)? _onUpdated;
  FutureOr<void> Function($onChangeParams)? _onRemoved;
 
  ${m.adapterName}({
    FutureOr<dynamic> Function($onRequestParams)? onRequest,
    FutureOr<void> Function($onResultsParams)? onResults,
    FutureOr<void> Function($onChangeParams)? onCreated,
    FutureOr<void> Function($onUpdateParams)? onUpdated,
    FutureOr<void> Function($onChangeParams)? onRemoved,
  }) :
    _onRequest = onRequest,
    _onResults = onResults,
    _onCreated = onCreated,
    _onUpdated = onUpdated,
    _onRemoved = onRemoved;

  FutureOr<dynamic> onRequest($onRequestParams) => _onRequest?.call($onRequestArgs);
  FutureOr<void> onResults($onResultsParams) => _onResults?.call($onResultsArgs);
  FutureOr<void> onCreated($onChangeParams) => _onCreated?.call($onChangeArgs);
  FutureOr<void> onUpdated($onUpdateParams) => _onUpdated?.call($onUpdateArgs);
  FutureOr<void> onRemoved($onChangeParams) => _onRemoved?.call($onChangeArgs);
  
  FutureOr<void> _created(Executor ex, $_onChangeParams) => onCreated($onChangeArgs);
  FutureOr<void> _updated(Executor ex, $_onUpdateParams) => onUpdated($onUpdateArgs);
  FutureOr<void> _removed(Executor ex, $_onChangeParams) => onRemoved($onChangeArgs);

  void expire([${m.queryName}? query]) {
    if(query == null || query.isAll) {
      _fetched.clear();
    } else {
      _fetched.remove(query.hashCode);
    }
  }
  
  final _fetched = <int, Future<void>>{};
  Future<void> fetch($fetchParams) {
    if (force) {
      final current = _fetched.remove(query.hashCode);
      if(current != null) {
        return current.then((_) => fetch($fetchArgs));
      }
    }
    return _fetched.putIfAbsent(query.hashCode, () => _fetch($_fetchArgs));
  }
  Future<void> _fetch($_fetchParams) async {
    _log.debug(() => 'fetching ${m.name}(\$query)');
    final data = await onRequest($onRequestArgs);
    if(data == null) {
      _log.debug(() => 'fetched ${m.name}: not found');
    }
    else if(data is Map<String, dynamic>) {
      await onResults($onResultsArgs);
      final ${m.varName} = ${m.name}.fromJson($jsonArgs);
      _feature.put(${m.varName});
      _log.debug(() => 'fetched \$${m.varName}');
    }
    else if(data is Iterable) {
      await onResults($onResultsArgs);
      final ${m.varNamePlural} = ${m.name}.fromJsonList($jsonArgs);
      _feature.set<${m.name}>($setArgs);
      _log.debug(() => 'fetched \$${m.varNamePlural}');
    }
    else {
      _log.debug(() => 'fetched ${m.name}: invalid result(\$data)');
    }
  }
}
''';

  String get onRequestParams => [
    '${schema.dsName} ds',
    if(m.isNotSingleton) '${m.queryName} query'
  ].join(',');

  String get onRequestArgs => [
    '_modelContext',
    if(m.isNotSingleton) 'query'
  ].join(',');

  String get onResultsParams => [
    '${schema.dsName} ds',
    if(m.isNotSingleton) '${m.queryName} query',
    if(m.isSingleton) 'Map<String, dynamic> data'
    else 'dynamic data'
  ].join(',');

  String get onResultsArgs => [
    '_modelContext',
    if(m.isNotSingleton) 'query',
    'data'
  ].join(',');

  String get onChangeParams => [
    '${schema.dsName} ds',
    '${m.name} ${m.varName}'
  ].join(',');

  String get _onChangeParams => [
    '${m.name} ${m.varName}'
  ].join(',');

  String get onChangeArgs => [
    '_modelContext',
    m.varName
  ].join(',');

  String get onUpdateParams => [
    '${schema.dsName} ds',
    '${m.name} old',
    '${m.name} ${m.varName}'
  ].join(',');

  String get _onUpdateParams => [
    '${m.name} old',
    '${m.name} ${m.varName}'
  ].join(',');

  String get onUpdateArgs => [
    '_modelContext',
    'old',
    m.varName
  ].join(',');

  String get fetchParams => [
    if(m.isNotSingleton) '${m.queryName} query',
    'bool force'
  ].join(',');

  String get fetchArgs => [
    if(m.isNotSingleton) 'query',
    'false'
  ].join(',');

  String get _fetchParams => [
    if(m.isNotSingleton) '${m.queryName} query',
  ].join(',');

  String get _fetchArgs => [
    if(m.isNotSingleton) 'query',
  ].join(',');

  String get jsonArgs => [
    'data',
    if(m.isNotSingleton) 'withQuery: query'
  ].join(',');

  String get setArgs => [
    m.varNamePlural,
    'where: query.test',
  ].join(',');
}

